procedure imclean(image,outimg,maskimg,regions)

string	image	=""		{prompt="Input image"}
string	outimg	=""		{prompt="Output cleaned image"}
string	maskimg	=""		{prompt="Input mask image (if none: \"\")"}
string	regions	=".pos.del .pos.cosmic"	{prompt="Space separated list of region files"}
real	badval	=0.		{prompt="Bad pixel value in mask image"}
real	goodval	=1.		{prompt="Good pixel value in mask image"}
bool	dotransp=no		{prompt="Transpose region coordinates?"}
bool	creamask=yes		{prompt="Create and save mask image?"}
string	outmask	=".mask"	{prompt="Ouput mask image name"}
bool	verbose	=yes		{prompt="Verbose messages?"}

# @(#) task imclean              Author: R.A. Jansen -- November 12 1996
# @(#)
# @(#) task to clean images by interpolation over bad pixels. This task
# @(#) serves as a user interface to executable 'imclean.exe'. Bad pixel
# @(#) regions should be given thru a mask image, pixel region files, or
# @(#) both. If no mask image is given and 'creamask'=yes a mask image
# @(#) will be created; if a mask image was given and 'outmask=maskimg',
# @(#) the input maskimage will be updated. Input and output images can
# @(#) have identical names.
# @(#) See also: imclean.f
# @(#)
# @(#) Jun 21 2000 -- If the input image has a format other than OIF, it
# @(#)                is copied to OIF, cleaned, and the result exported
# @(#)                back to an image of the input image format. (RAJ).

begin
	string	img, oimg, mask, omask, reg, imtyp, simg
	string	file1, file2, file3, cline, blank
	struct  line
	real	vbad, vgood
	int	ncol, nrow
	bool	tmpcopy

	img	= image
	oimg	= outimg
	mask	= maskimg
	omask	= outmask
	vbad	= badval
	vgood	= goodval
	reg	= regions

	file1	= ""
	file2	= ""
	file3	= ""
	blank	= "                                        "

	if (verbose) { print (" ") }

#	Check consistency of the input...
	if ( img == "" )  {
	   error (1, "No input image given!")
	}
	if ( oimg == "" ) {
	   error (1, "No output (cleaned) image specified!")
	}
	if ( mask == "" && reg == "" ) {
	   error (1, "Neither input mask image nor region files specified!")
	}
	if ( creamask && omask == "" ) {
	   error (1, "'Creamask=yes', while 'outmask==\"\"'!")
	}

#	Print opening message...
	gdate() ; sysinfo()
	print ("IMCLEAN:   "//sysinfo.sysmsg//"   "//gdate.tdate)

#       Check existence input, output and mask images...
	unlearn chkimg
	chkimg (img, "access", verbose=yes)
	if (!chkimg.ok) { return }
	imtyp = chkimg.imtype
	simg  = chkimg.imroot
	if ( imtyp != ".imh" || chkimg.imrpstr != "" ) {
#	   work on a copy of (the relevant part of) the input image...
	   print (" Creating .imh working copy of "//imtyp//" format input image:")
	   print (" "//simg//imtyp//chkimg.imrpstr//" --> tmp"//chkimg.imroot//".imh ...")
	   imcopy (simg//imtyp//chkimg.imrpstr, "tmp"//chkimg.imroot//".imh", verbose=no)
	   img = "tmp"//chkimg.imroot//".imh"
	   tmpcopy = yes
	} else {
#	   work on the original...
	   img = simg//imtyp
	   tmpcopy = no
	}

	imgets (img, "naxis1") ; nrow = int(imgets.value)
	imgets (img, "naxis1") ; nrow = int(imgets.value)
	imgets (img, "naxis2") ; ncol = int(imgets.value)

	chkimg (oimg, "access", verbose=no)
	if (chkimg.ok && chkimg.imroot//chkimg.imtype != simg//imtyp ) {
	   error (1, "Output cleaned image "//oimg//" already exists!")
	}
	oimg = chkimg.imroot//chkimg.imtype

        if ( mask != "" && substr(mask,1,1) == "." ) { mask = simg//mask }
        if ( omask != "" && substr(omask,1,1) == "." ) { omask = simg//omask }
	if ( mask != "" ) {
	   chkimg (mask, "access", verbose=no)
	   if (!chkimg.ok) { error (1, "Input mask image "//mask//" not found!") }
	   if (tmpcopy) {
	      print (" Creating .imh working copy of "//imtyp//" format mask image:")
	      print (" "//mask//" --> tmp"//mask//".imh ...")
	      imcopy (mask, "tmp"//mask//".imh", verbose=no)
	   }
	}

	if (creamask) {
	   chkimg (omask, "access", verbose=no)
	   omask = chkimg.imroot
	} else {
	   omask = ""
	}

#       Parse the regions string and check existence region files...
        ocparse (reg, delim=" ")
        if ( ocparse.nfields > 3 ) {
           error (1, "Current maximum to number of region files is 3! Exit...")
        } else {
           if ( ocparse.nfields >= 1 ) { 
              file1 = ocparse.field1
              if ( substr(file1,1,1) == "." ) { file1 = simg//file1 }
              if ( !access(file1) ) {
                 #print ("Warning: region file "//file1//" not found!")
                 file1 = ""
              }
           }
           if ( ocparse.nfields >= 2 ) {
              file2 = ocparse.field2
              if ( substr(file2,1,1) == "." ) { file2 = simg//file2 }
              if ( !access(file2) ) {
                 #print ("Warning: region file "//file2//" not found!")
                 file2 = ""
              }
           }
           if ( ocparse.nfields == 3 ) {
              file3 = ocparse.field3
              if ( substr(file2,1,1) == "." ) { file3 = simg//file3 }
              if ( !access(file3) ) {
                 #print ("Warning: region file "//file3//" not found!")
                 file3 = ""
              }
           }
        }
	if ( file1 == "" && file2 == "" && file3 == "" ) {
	   print ("Warning: Could not find any of the region files!")
	}

#       Print input summary now everything seems to be OK...
	if (verbose) {
	   print ("----------------------   i m c l e a n . e x e   -----------------------"
	   print ("   img    = "//img)
	   print ("   output = "//oimg)
	   if ( tmpcopy && mask != "" ) { print ("   inmask = tmp"//mask//".imh") }
	   if ( !tmpcopy && mask != "") { print ("   inmask = "//mask//imtyp) }
	   if ( omask != "" ) { print ("   outmask= "//omask//imtyp) }
	   cline = "   regions= "//file1
	   if (file2 != "" ) { cline = cline//","//file2 }
	   if (file3 != "" ) { cline = cline//","//file3 }
	   print (cline)
           print ("   badval =",badval,"  goodval=",goodval,
		  "  creamask=",creamask,"   dotransp=",dotransp)
	}

#       Prepare input text file
        print (img, > "tmpmcln.inp")
	if (mask == "" ) { print ("", >> "tmpmcln.inp") }
	if (tmpcopy && mask != "" ) { print ("tmp"//mask//".imh", >> "tmpmcln.inp") }
	if (!tmpcopy && mask != "" ) { print (mask//imtyp, >> "tmpmcln.inp") }
        print (vbad, >> "tmpmcln.inp")
        print (vgood, >> "tmpmcln.inp")
        print (file1, >> "tmpmcln.inp")
        print (file2, >> "tmpmcln.inp")
        print (file3, >> "tmpmcln.inp")
        print (creamask, >> "tmpmcln.inp")
        print (dotransp, >> "tmpmcln.inp")

#       Make sure executable is present in working directory...
	if (ncol<=999&&nrow<=999) {
	   if ( !access("imclean999.exe") ) {
	      copy ("imclean$imclean999.exe", ".", verbose=no) 
	   }
	   if ( access("imclean2064.exe") ) {
	      delete ("imclean2064.exe", yes, verify=no)
	   }
	} else {
	   if ( access("imclean999.exe") ) {
	      delete ("imclean999.exe", yes, verify=no)
	   }
	   if ( !access("imclean2064.exe") ) {
	      copy ("imclean$imclean2064.exe", ".", verbose=no) 
	   }
	}

#       Execute imclean.exe ...
	if (ncol<=999&&nrow<=999) {
	   print ("!imclean999.exe < tmpmcln.inp", > "tmpmcln.cl")
	} else {
	   print ("!imclean2064.exe < tmpmcln.inp", > "tmpmcln.cl")
	}
	cl < tmpmcln.cl

	if (verbose) {
	   print ("------------------------------------------------------------------------"
	}


#       Rename to proper output names; default extensions in 'imclean.exe'
#       are 'c' for the cleaned frame and either '.mask' (if no input mask
#       frame was given or if the output mask frame name differs from that of
#       the input mask) or '.masknew' (if the input and output masks have the
#       same name).  Clean up the temporary files.
        if (access("imclean999.exe")) { delete("imclean999.exe", yes, verify=no) }
        if (access("imclean2064.exe")) { delete("imclean2064.exe", yes, verify=no) }
	delete ("tmpmcln.*", yes, verify=no)
	if ( oimg == simg//imtyp ) {
	   imdelete (simg//imtyp, yes, verify=no)
	}
	if (tmpcopy) {
	   imdelete (img, yes, verify=no)
           if (verbose) { print (" Renaming tmp"//simg//"c.imh --> "//oimg//" ...") }
	   imcopy ("tmp"//simg//"c.imh", oimg, verbose=no)
	   imdelete ("tmp"//simg//"c.imh", yes, verify=no)
	   if (creamask) {
	      if (mask != "") { imdelete ("tmp"//mask//".imh", yes, verify=no) }
              if (mask == omask) {
		 imdelete (mask//imtyp, yes, verify=no)
		 if (verbose) { print (" Renaming tmp"//simg//".masknew.imh --> "//omask//imtyp//" ...") }
		 imcopy ("tmp"//simg//".masknew.imh", omask//imtyp, verbose=no)
		 imdelete ("tmp"//simg//".masknew.imh", yes, verify=no)
              } else {
                 if (verbose) { print (" Renaming tmp"//simg//".mask.imh --> "//omask//imtyp//" ...") }
		 imcopy ("tmp"//simg//".mask.imh", omask//imtyp, verbose=no)
		 imdelete ("tmp"//simg//".mask.imh", yes, verify=no)
	      }
	   }
	} else {
	   if ( oimg != simg//"c"//imtyp ) {
              if (verbose) { print (" Renaming "//simg//"c"//imtyp//" --> "//oimg//" ...") }
	      imcopy (simg//"c"//imtyp, oimg, verbose=no)
	      imdelete (simg//"c"//imtyp, yes, verify=no)
	   }
	   if (creamask) {
              if (mask == omask) {
		 imdelete (mask//imtyp, yes, verify=no)
		 if (verbose) { print (" Renaming "//omask//"new"//imtyp//" --> "//omask//imtyp//" ...") }
		 imcopy (omask//"new"//imtyp, omask//imtyp, verbose=no)
		 imdelete (omask//"new"//imtyp, yes, verify=no)
	      }
	   }
	}

end
