      PROGRAM imclean

C @(#)Program: imclean.f             Author: R.A. Jansen -- Nov 11, 1996
C @(#)
C @(#)Program to interpolate over circular and/or square pixel regions
C @(#)in an input image, using either a mask image or a (set of) region
C @(#)files. The pixel regions are to be given in region files, listing
C @(#)one region per line. Regions can either be square or circular and
C @(#)should be given in IRAF 'tvmark' readable format.
C @(#)The mask image (that imclean produces for internal use when region
C @(#)files are used) can be saved.
C @(#)
C @(#)To compile, start up iraf and type: fc imclean.f -o imclean.exe
C @(#)
C @(#)Includes subroutines: imglen,GETINP,IMREAD,CREAMASK,CPARSE,IMWRITE
C @(#)                      

      CHARACTER*80 image,mimage,pstar,pdel,pcos,cext*24
      REAL    pix(2064,2064),mask(2064,2064)
      REAL    maskv,goodv,lastgood,nextgood,lw,nw,tw
      INTEGER ierr,imglen,nrow,ncol,mrow,mcol,i,j,k,m
      LOGICAL savem,inmask,transp

      DATA pix/4260096*0./,mask/4260096*0./

C ----------------------      M   A   I   N      -----------------------

C     Initialize
      ierr=0
      savem=.FALSE.
      inmask=.FALSE.
      transp=.FALSE.

C     Get input
      CALL GETINP(image,mimage,maskv,goodv,pstar,pdel,pcos,
     +            savem,transp,ierr)
      IF (ierr.NE.0) GOTO 98
      IF (mimage(1:imglen(mimage)).NE.'none') inmask = .TRUE.

C     Read input image 'image' into pixel array pix(i,j)
      CALL IMREAD(image,pix,nrow,ncol,ierr)
      IF (ierr.NE.0) GOTO 98

C     If a mask image is defined, read in mask image
      IF (inmask) THEN
         CALL IMREAD(mimage,mask,mrow,mcol,ierr)
         IF (ierr.NE.0) GOTO 98
         IF (mrow.NE.nrow.OR.mcol.NE.ncol) THEN
            WRITE(*,'(''ERROR: image and mask image differ in size!'')')
            GOTO 99
         ENDIF
C        Replace 'maskv' by -33000. and other pixels by 0. in mask
         DO j=1,nrow
            DO i=1,ncol
               IF (mask(i,j).EQ.maskv) THEN
                  mask(i,j) = -33000.
               ELSE
                  mask(i,j) = 0.
               ENDIF
            ENDDO
         ENDDO
      ENDIF

C     Replace all pixels of the mask array in regions affected by stars,
C     cosmics or CCD defects (from region files) with value -33000.
      IF (pstar(1:imglen(pstar)).EQ.'none') GOTO 10
5     OPEN(UNIT=25,FILE=pstar(1:imglen(pstar)),STATUS='OLD',ERR=10)
      CLOSE(UNIT=25)
      CALL CREAMASK(mask,nrow,ncol,pstar,transp)
10    IF (pcos(1:imglen(pcos)).EQ.'none') GOTO 15
      OPEN(UNIT=25,FILE=pcos(1:imglen(pcos)),STATUS='OLD',ERR=15)
      CLOSE(UNIT=25)
      CALL CREAMASK(mask,nrow,ncol,pcos,transp)
15    IF (pdel(1:imglen(pdel)).EQ.'none') GOTO 20
      OPEN(UNIT=25,FILE=pdel(1:imglen(pdel)),STATUS='OLD',ERR=20)
      CLOSE(UNIT=25)
      CALL CREAMASK(mask,nrow,ncol,pdel,transp)
20    CONTINUE

C     Now we have a image array 'pix' and a mask array 'mask'.
C     Loop over all pixels line per line and keep track of the last
C     'good' pixel. If a 'bad' pixel is encountered, find the first
C     pixel that is good again.
      lastgood = -33000.
      DO j=1,nrow
         DO i=1,ncol-1
            k = i
            IF (mask(i,j).NE.-33000.) THEN
               lastgood = pix(i,j)
               m = i
            ELSE
C              Loop over mimimum number of pixels to get nextgood value
               nextgood = -33000.
25             k = k + 1
               IF (mask(k,j).NE.-33000..OR.k.EQ.ncol) THEN
                  IF (mask(k,j).NE.-33000.) nextgood = pix(k,j)
                  CONTINUE
               ELSE
                  GOTO 25
               ENDIF
C              We now have found either the last and next 'good' pixels
C              or both or either of them is -33000. (bad pixels extend
C              to edge of frame).
               IF (lastgood.NE.-33000..AND.nextgood.NE.-33000.) THEN
                  lw = 1.0/((i-m)*(i-m))
                  nw = 1.0/((k-i)*(k-i))
                  tw = 1.0/(lw+nw)
                  pix(i,j) = tw*(lw*lastgood + nw*nextgood)
               ELSE IF (lastgood.NE.-33000.) THEN
                  pix(i,j) = lastgood
               ELSE IF (nextgood.NE.-33000.) THEN
                  pix(i,j) = nextgood
               ELSE
                  pix(i,j) = -33000.
               ENDIF
            ENDIF
         ENDDO
      ENDDO

C     We have now interpolated across (most of the) bad pixels along
C     the lines; next repeat along the columns, but use all interpolated
C     'good' values from the first interpolation pass in the calculation
C     of the new output values as well.
      DO i=1,ncol
         DO j=1,nrow-1
            k = j
            IF (mask(i,j).NE.-33000.) THEN
               lastgood = pix(i,j)
               m = j
            ELSE
C              Loop over mimimum number of pixels to get nextgood value
               nextgood = -33000.
30             k = k + 1
               IF (mask(i,k).NE.-33000..OR.k.EQ.nrow) THEN
                  IF (mask(i,k).NE.-33000.) nextgood = pix(i,k)
                  CONTINUE
               ELSE
                  GOTO 30
               ENDIF
C              We now have found either the last and next 'good' pixels
C              or both or either of them is -33000. (bad pixels extend
C              to edge of frame).
               IF (lastgood.NE.-33000..AND.nextgood.NE.-33000.) THEN
                  lw = 1.0/((j-m)*(j-m))
                  nw = 1.0/((k-j)*(k-j))
                  tw = 1.0/(lw+nw)
                  IF (pix(i,j).NE.-33000.) THEN
                    pix(i,j)=0.5*(pix(i,j)+tw*(lw*lastgood+nw*nextgood))
                  ELSE
                    pix(i,j)=tw*(lw*lastgood+nw*nextgood)
                  ENDIF
               ELSE IF (lastgood.NE.-33000.) THEN
                  IF (pix(i,j).NE.-33000.) THEN
                     pix(i,j) = 0.5*(pix(i,j)+lastgood)
                  ELSE
                     pix(i,j) = lastgood
                  ENDIF
               ELSE IF (nextgood.NE.-33000.) THEN
                  IF (pix(i,j).NE.-33000.) THEN
                     pix(i,j) = 0.5*(pix(i,j)+nextgood)
                  ELSE
                     pix(i,j) = nextgood
                  ENDIF
               ELSE
C                 Pixel value stays either 'bad' if it is positioned
C                 at the crossing of a completely bad column and
C                 completely bad row, or stays at the value determined
C                 in the first interpolation pass; i.e. pix(i,j)=pix(i,j)
                  CONTINUE
               ENDIF
            ENDIF
         ENDDO
      ENDDO

C     Write pixel array pix(i,j) into image
      cext = 'c                       '
      CALL IMWRITE(image,pix,nrow,ncol,cext,ierr)
      IF (ierr.NE.0) GOTO 98

C     If savem=true, write out mask frame
      IF (savem.EQ..TRUE.) THEN
C        Replace -33000. by 'maskv' and other pixels by 0.
         DO j=1,nrow
            DO i=1,ncol
               IF (mask(i,j).EQ.-33000.) THEN
                  mask(i,j) = maskv
               ELSE
                  mask(i,j) = goodv
               ENDIF
            ENDDO
         ENDDO
         IF (inmask) THEN
            cext = 'new                     '
            CALL IMWRITE(mimage,mask,nrow,ncol,cext,ierr)
         ELSE
            cext = '.mask                   '
            CALL IMWRITE(image,mask,nrow,ncol,cext,ierr)
         ENDIF
         IF (ierr.NE.0) GOTO 98
      ENDIF

C     End message
C     WRITE(*,'(''IMCLEAN: Finished.'')')
      GOTO 99

C     Error handling
98    IF (ierr.EQ.99999) THEN
         WRITE(*,'(''ERROR: image is larger than 2064 x 2064 pixels'')')
      ELSE
         WRITE(*,'(''ERROR: Somewhere...'')')
      ENDIF

99    STOP
      END


C ----------------------------------------------------------------------

      FUNCTION imglen(cstring)
         
      CHARACTER*80 cstring
      INTEGER imglen,i

      DO 1 i=LEN(cstring),1,-1
         imglen=i
         IF (cstring(imglen:imglen).NE.'') GOTO 2
1     CONTINUE
      imglen=0
      RETURN

2     IF (cstring(imglen-3:imglen).EQ.'.imh') imglen = imglen-4

      RETURN
      END   

C ----------------------------------------------------------------------

      SUBROUTINE GETINP(image,mimage,maskval,goodval,pstar,pdel,pcos,
     +                  savem,transp,ierr)

      CHARACTER*80 image,mimage,pstar,pdel,pcos,csavem,ctransp,cstr
      REAL    maskval,goodval
      INTEGER ierr
      LOGICAL savem,transp

C     Initialize
      ierr    = 0
      goodval = 0.
      maskval = -33000.
      savem   = .FALSE.
      transp  = .FALSE.
      pcos    = 'none'
      pdel    = 'none'
      pstar   = 'none'
      image   = ' '
      mimage  = 'none'

C     NOTE: this program is suposed to be called via IRAF task
C           imclean.cl; therefore no prompting for input should be
C           required. For stand-alone use, uncomment the following
C           'write' statements

C     WRITE(*,*)'Image (\"\"):'
      READ(*,'(A80)') image
C     WRITE(*,*)'Mask image (\"\"):'
      READ(*,'(A80)') mimage
C     WRITE(*,*)'Mask value (bad pixels) (-33000.):'
      READ(*,'(A80)') cstr
      IF (cstr.NE.'') READ(cstr,*) maskval
C     WRITE(*,*)'Good value (good pixels) (0.):'
      READ(*,'(A80)') cstr
      IF (cstr.NE.'') READ(cstr,*) goodval
C     WRITE(*,*)'Region file 1 (\"\"):'
      READ(*,'(A80)') pstar
C     WRITE(*,*)'Region file 2 (\"\"):' 
      READ(*,'(A80)') pdel
C     WRITE(*,*)'Region file 3 (\"\"):' 
      READ(*,'(A80)') pcos
C     WRITE(*,*)'Save mask image? (no):'
      READ(*,'(A80)') csavem
      IF (csavem(1:1).EQ.'y') savem = .TRUE.
      IF (csavem(1:4).EQ.'true') savem = .TRUE.
      IF (csavem(1:6).EQ.'.true.') savem = .TRUE.
C     WRITE(*,*)'Transpose region coordinates? (no):'
      READ(*,'(A80)') ctransp
      IF (ctransp(1:1).EQ.'y') transp = .TRUE.
      IF (ctransp(1:4).EQ.'true') transp = .TRUE.
      IF (ctransp(1:6).EQ.'.true.') transp = .TRUE.

      IF (mimage.EQ.'') mimage = 'none'
      IF (pstar.EQ.'')  pstar  = 'none'
      IF (pcos.EQ.'')   pcos  = 'none'
      IF (pdel.EQ.'')   pdel  = 'none'

9     RETURN
      END


C ----------------------------------------------------------------------

      SUBROUTINE IMREAD(image,pix,nrow,ncol,ierr)

      CHARACTER image*80,imgstr*80
      INTEGER nrow,ncol,ierr,img1,axlen(7),naxis,dtype,imglen,i
      REAL tmppix(2064),pix(2064,2064)
      DATA tmppix/2064*0./

      ierr=0
      DO i=1,80
         imgstr(i:i)=' '
      ENDDO
C      imgstr = image(1:imglen(image))

C     Open input IRAF image
C      CALL imopen(imgstr,1,img1,ierr)
      CALL imopen(image,1,img1,ierr)
      IF (ierr.NE.0) GOTO 9

C     Determine size of image (implicit 2-D image)
      CALL imgsiz(img1,axlen,naxis,dtype,ierr)
      IF (ierr.NE.0) GOTO 9
      ncol = axlen(1)
      nrow = axlen(2)

C     Read IRAF pixel file into array pix(i,j) via temporary image line
C     buffer tmppix(i) (real).
      IF (dtype.EQ.3) WRITE(*,'((A),''['',I4,'','',I4,''][short]...'')')
     +                ' Reading '//image(1:imglen(image)),ncol,nrow
      IF (dtype.EQ.6) WRITE(*,'((A),''['',I4,'','',I4,''][real] ...'')')
     +                ' Reading '//image(1:imglen(image)),ncol,nrow
      IF (ncol.GT.2064.OR.nrow.GT.2064) ierr=99999  
      IF (ierr.NE.0) GOTO 9

C     Read IRAF pixel file into array pix(i,j) via temporary image line
C     buffer tmppix(i)
      DO j=1,nrow
         CALL imgl2r(img1,tmppix,j,ierr)
         IF (ierr.NE.0) GOTO 9
         DO i=1,ncol
            pix(i,j) = tmppix(i)
         ENDDO
      ENDDO

C     Clean up
      CALL imclos(img1,ierr)

9     RETURN
      END


C ----------------------------------------------------------------------
      SUBROUTINE IMWRITE(image,newpix,nrow,ncol,cext,ierr)

      CHARACTER image*80,newimage*80,cext*24
      INTEGER nrow,ncol,ierr,img1,img2,i,j
      REAL newpix(2064,2064),tmppix(2064)
      DATA tmppix/2064*0./

      ierr = 0
      newimage=' '

C     Determine output image name
      newimage = image(1:imglen(image))//cext

C     Open input image and create a new-copy output image
      CALL imopen(image(1:imglen(image)),1,img1,ierr)
      IF (ierr.NE.0) GOTO 9
      CALL imopnc(newimage(1:imglen(newimage)),img1,img2,ierr)
      IF (ierr.NE.0) GOTO 9

C     Write out new IRAF pixel file
      WRITE(*,'('' Writing '',(A),''['',I4,'','',I4,''][real]...'')') 
     + newimage(1:imglen(newimage)),ncol,nrow
      DO j=1,nrow
         DO i=1,ncol
            tmppix(i)=newpix(i,j)
         ENDDO
         CALL impl2r(img2,tmppix,j,ierr)
         IF (ierr.NE.0) GOTO 9
      ENDDO

C     Clean up
      CALL imclos(img1,ierr)
      IF (ierr.NE.0) GOTO 9
      CALL imclos(img2,ierr)

9     RETURN
      END


C ----------------------------------------------------------------------

      SUBROUTINE CREAMASK(mask,nrow,ncol,posfil,transp)
            
      CHARACTER*80 posfil,cstr,sep*1
      REAL mask(2064,2064),f1,f2,f3,f4,rtmp
      INTEGER nrow,ncol,i,j,imglen,nf,ierr
      LOGICAL transp

      ierr = 0
      sep = ' '

C     We allready checked in MAIN whether 'posfil' exists
      OPEN(UNIT=25,FILE=posfil(1:imglen(posfil)),STATUS='OLD')
      WRITE(*,'('' Reading bad pixel regions file '',(A),'' ...'')')
     +                                       posfil(1:imglen(posfil))
1     READ(25,'(A80)',END=9) cstr
      CALL CPARSE(cstr,sep,nf,f1,f2,f3,f4,ierr)

      IF (nf.EQ.4) THEN
        IF (f1.GT.f2) THEN
           rtmp = f1
           f1 = f2
           f2 = rtmp
        ENDIF
        IF (f3.GT.f4) THEN
           rtmp = f3
           f3 = f4
           f4 = rtmp
        ENDIF
        IF (transp) THEN
           itmp1 = nrow - f4 + 1
           itmp2 = nrow - f3 + 1
           itmp3 = ncol - f2 + 1
           itmp4 = ncol - f1 + 1
           DO i=NINT(MAX(1.,itmp1)),NINT(MIN(itmp2,REAL(ncol))),1
              DO j=NINT(MAX(1.,itmp3)),NINT(MIN(itmp4,REAL(ncol))),1
                 mask(i,j)=-33000.
              ENDDO
           ENDDO
        ELSE  
           DO i=NINT(MAX(1.,f1)),NINT(MIN(f2,REAL(ncol))),1
              DO j=NINT(MAX(1.,f3)),NINT(MIN(f4,REAL(nrow))),1
                 mask(i,j)=-33000.
              ENDDO
           ENDDO
        ENDIF
      ELSE IF (nf.EQ.3) THEN
        IF (transp) THEN
           itmp1 = nrow - f2 + 1 - f3
           itmp2 = nrow - f2 + 1 + f3
           itmp3 = ncol - f1 + 1 - f3
           itmp4 = ncol - f1 + 1 + f3
           DO i=NINT(MAX(1.,itmp1)),NINT(MIN(itmp2,REAL(ncol))),1
              DO j=NINT(MAX(1.,itmp3)),NINT(min(itmp4,REAL(nrow))),1
                 rtmp = (i-nrow+f2-1)*(i-nrow+f2-1)
                 rtmp = rtmp + (j-ncol+f1-1)*(j-ncol+f1-1)
                 IF ( sqrt(rtmp).LE.f3 ) mask(i,j)=-33000.
              ENDDO
           ENDDO
        ELSE
           DO i=NINT(MAX(1.,(f1-f3))),NINT(MIN((f1+f3),REAL(ncol))),1
              DO j=NINT(MAX(1.,(f2-f3))),NINT(MIN((f2+f3),REAL(nrow))),1
                 rtmp = (i-f1)*(i-f1) + (j-f2)*(j-f2)
                 IF ( sqrt(rtmp).LE.f3 ) mask(i,j)=-33000.
              ENDDO
           ENDDO
        ENDIF
      ELSE
        WRITE(*,'(''ERROR: Skipping unrecognized maskregion '',
     +                             (A),''...'')') cstr(1:imglen(cstr))
      ENDIF
      GOTO 1 
      
9     CLOSE(UNIT=25)
      RETURN
      END


C ----------------------------------------------------------------------

      SUBROUTINE CPARSE(cline,delim,nf,f1,f2,f3,f4,ierr)

      CHARACTER cline*80,delim*1,cchar*1,ccharold*1,ftmp*80
      REAL      f1,f2,f3,f4
      INTEGER   i,nf,imglen,nchar,ierr,count
          
      ierr=0 
      cchar=' '
      ccharold=cchar
      ftmp=' '
      f1=0.00
      f2=0.00
      f3=0.00
      f4=0.00
      nchar=imglen(cline) 

      nf=0
      count=0
      DO i=1,nchar,1
         cchar=cline(i:i)
         IF (cchar.NE.delim.AND.i.NE.nchar) THEN
            count = count + 1
            ftmp(count:count) = cchar
         ELSE IF (cchar.EQ.delim.AND.ccharold.EQ.delim) THEN
C           ignore leading and multiple 'separators'
            CONTINUE
         ELSE IF (cchar.EQ.delim.OR.i.EQ.nchar) THEN
            IF (i.EQ.nchar.AND.cchar.NE.delim) THEN
               count = count + 1
               ftmp(count:count) = cchar
            ENDIF
            nf=nf+1
            IF (nf.EQ.1) READ(ftmp,*) f1
            IF (nf.EQ.2) READ(ftmp,*) f2
            IF (nf.EQ.3) READ(ftmp,*) f3
            IF (nf.EQ.4) READ(ftmp,*) f4
            ftmp=' '
            count = 0
         ELSE
C           It should be impossible to get here!
            ierr=1
            WRITE(*,'(''ERROR: error in CPARSE somewhere!'')')
         ENDIF   
         ccharold = cchar
      ENDDO
            
            
      RETURN
      END
         

